%%%%%%%%%%%%%%%%%%%
% Script to deblur an image given the Piecewise Rigid Scene Flow computed
% between stereo video frames
% as described in the paper: 
% Sellent, Rother, Roth: "Stereo Video Deblurring",
% In: ECCV 2016 (to appear)
%
% Implementation by Anita Sellent (anita.sellent@tu-dresden.de)
% 
% Computer Vision Lab Dresden
% Technische Universitaet Dresden, Germany
% 
% Last revision: July 2016
%
% This program can be used only for research purposes.
% This program is distributed WITHOUT ANY WARRANTY; 
% without even the implied warranty of MERCHANTABILITY 
% or FITNESS FOR A PARTICULAR PURPOSE.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
function sharp = deconvL2_wSegmentColor( L1_big, ...
    validL1, ...
    KL1, ...
    wk, hk, ...
    weight_x, weight_y, weight_xx, weight_yy, weight_xy, ...
    parameter)

%deconvolve images using the conjugate gradient method
%this code is of the basis of Levin's implementation available here: http://groups.csail.mit.edu/graphics/CodedAperture/
%thanks!

%% constants
doVisualize = parameter.doVisualize;
alpha = parameter.alpha;
nIter = parameter.nIter;

eps = 1e-10;%accuracy for gradient descent. Set very small to run a fixed number of iterations

%% initializations
[ height, width, colors ] = size( L1_big );

%derivative filter
dxf = [1 -1];
dyf = [1;-1];
dyyf = [-1; 2; -1];
dxxf = [-1, 2, -1];
dxyf = [-1 1;1 -1];

%initialize the sharp image
x = L1_big;

if doVisualize
    x0 = x;
end

%% validity considerations
maskL1 = repmat( validL1, [1,1,colors] );

%% weights
[ ~, ~, channels ] = size( weight_x );
if channels < colors
    weight_x = repmat( weight_x(:,:,1), [ 1, 1, colors] );
    weight_y = repmat( weight_y(:,:,1), [ 1, 1, colors] );
    weight_xx = repmat( weight_xx(:,:,1), [ 1, 1, colors] );
    weight_xy = repmat( weight_xy(:,:,1), [ 1, 1, colors] );
    weight_yy = repmat( weight_yy(:,:,1), [ 1, 1, colors] );
end

%% gradient descent
b = applyKernelMatrix( L1_big.*maskL1, KL1' );

%data term
Ax = applyKernelMatrix( ...
    applyKernelMatrix( x, KL1, height, width, colors) ...
    .* maskL1, KL1' );

%regularization 
Ax = Ax + alpha * conv2Color( weight_x.* ...
    conv2Color( x, rot90( dxf, 2 ),'valid'), dxf );
Ax = Ax + alpha * conv2Color( weight_y.*...
    conv2Color( x, rot90( dyf, 2 ),'valid'), dyf);
Ax = Ax + alpha * conv2Color( weight_xx.*...
    conv2Color( x, rot90( dxxf, 2 ),'valid'), dxxf);
Ax = Ax + alpha * conv2Color( weight_yy.*...
    conv2Color( x, rot90( dyyf, 2 ),'valid'), dyyf);
Ax = Ax + alpha * conv2Color( weight_xy.*...
    conv2Color( x, rot90( dxyf, 2 ),'valid'), dxyf);

x = x(:);
r = b - Ax;
r = r(:);
p = r;
rsold = r' * r;

%iterate for a fixed max. number of iteration
for i=1:nIter
    if doVisualize
        fprintf( ' %d,', i )
    end
    
    %data term
    Ap = applyKernelMatrix( ...
	  applyKernelMatrix( p, KL1, height, width, colors) ...
	  .* maskL1, KL1' );
    
    %regularization
    Ap = Ap + alpha * conv2Color( weight_x.*...
	  conv2Color( reshape( p, height, width, colors ), rot90( dxf, 2 ), 'valid' ), ...
	  dxf, 'full');
    Ap = Ap + alpha * conv2Color( weight_y.*...
	  conv2Color( reshape( p, height, width, colors ), rot90( dyf, 2 ), 'valid' ), ...
	  dyf, 'full');
    Ap = Ap + alpha * conv2Color( weight_xx.*...
	  conv2Color( reshape( p, height, width, colors ), rot90( dxxf, 2 ), 'valid' ), ...
	  dxxf, 'full');
    Ap = Ap + alpha * conv2Color( weight_yy.*...
	  conv2Color( reshape( p, height, width, colors ), rot90( dyyf, 2 ), 'valid' ), ...
	  dyyf, 'full');
    Ap = Ap + alpha * conv2Color( weight_xy.*...
	  conv2Color( reshape( p, height, width, colors ), rot90( dxyf, 2 ), 'valid' ), ...
	  dxyf, 'full');
    
    Ap = Ap(:);
    
    a = rsold / ( p' * Ap );
    x = x + a * p;
    r = r - a * Ap;
    rsnew = r' * r;
    if  rsnew < (eps*eps)
        break;
    end
    p = r + rsnew / rsold * p;
    rsold = rsnew;
    
    if doVisualize
        subplot( 1,2,1); imagesc( reshape( x0, height, width, colors), [0,1] );
        title('input (boundary filled)')
        colormap gray
        
        tmp = min( 1, max( 0, x ) );
        subplot( 1,2,2); imagesc( reshape( tmp, height, width, colors), [0,1] );
        title('current deblurring')
        colormap gray
        drawnow();
    end
    
end
if doVisualize
    fprintf('\n' )
end
sharp = reshape( x, height, width, colors);
