%%%%%%%%%%%%%%%%%%%
% Script to deblur an image given the Piecewise Rigid Scene Flow computed
% between stereo video frames
% as described in the paper: 
% Sellent, Rother, Roth: "Stereo Video Deblurring",
% In: ECCV 2016 (to appear)
%
% Implementation by Anita Sellent (anita.sellent@tu-dresden.de)
% 
% Computer Vision Lab Dresden
% Technische Universitaet Dresden, Germany
% 
% Last revision: July 2016
%
% This program can be used only for research purposes.
% This program is distributed WITHOUT ANY WARRANTY; 
% without even the implied warranty of MERCHANTABILITY 
% or FITNESS FOR A PARTICULAR PURPOSE.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 
function sharp = deconvSpsSegmentColor( L1, ...
    validL1, ...
    KL1, wk, hk, ...
    boundL1, parameter )

%deconvolve by iterativly adapting weights for data-term and Laplacian prior adaptation

%% constants
doVisualize = parameter.doVisualize;
useValid = parameter.useValid;


w0 = 0.1;%see IRLS
exp_a = 0.8;%as in %rho( c ) = |c|^{0.8} after eq. 7
thr_e = 0.01;%eq. 8

validWeight = parameter.validWeight;
alterIters = parameter.alterIters;

%% initializations
[ height0, width0, colors ] = size( L1 );
%derivative filter
dxf = [1 -1];
dyf = [1;-1];
dyyf=[-1; 2; -1];
dxxf=[-1, 2, -1];
dxyf=[-1 1;1 -1];

%% a boundary extended image
width = width0 + 2*wk;
height = height0 + 2*hk;

%% the boundary increased image
tmpPSF = zeros( 2*hk+1, 2*wk+1 );
tmpPSF( hk+1, wk+1 ) = 1;
L1_big = imfilter( L1, tmpPSF, 'corr', 'full', 'replicate');

%% initialize dataterm weight with occlusion masks
validL1_big = double( boundL1 );
validL1_big( hk + ( 1: height0), wk + ( 1: width0) ) = validL1;

%% initialize regularization weights as 1
weight_x = ones( height, width - 1);
weight_y = ones( height - 1, width);
weight_xx = ones( height, width - 2);
weight_yy = ones( height - 2, width);
weight_xy = ones( height - 1, width - 1);

%initial deconvolution    
[x] = deconvL2_wSegmentColor( L1_big, validL1_big, KL1, wk, hk, ...
    weight_x, weight_y, weight_xx, weight_yy, weight_xy, parameter);

%% refine
if ( 0 == mod( alterIters, 2) )
    alterIters = alterIters + 1;%make sure last update is on regularization
end

for t=1:alterIters
	%
	if useValid && (0 == mod( t, 2) )
	    %update the weights for data term
	    residualL1_sq = mean( ( applyKernelMatrix( x, KL1 ) - L1_big ).^2, 3);
            ucL1 = exp( - validWeight * residualL1_sq);
        
	    %define the modified objective
	    validL1_big = ucL1;
	    validL1_big = validL1_big .* double( boundL1 );
        
	end
    
	if (1 == mod( t, 2) )
	    %%compute weights for regularization
	    dy = conv2Color( x, rot90( dyf, 2),'valid');
	    dx = conv2Color( x, rot90( dxf, 2),'valid');
	    dyy = conv2Color( x, rot90( dyyf, 2),'valid');
	    dxx = conv2Color( x, rot90( dxxf, 2),'valid');
	    dxy = conv2Color( x, rot90( dxyf, 2),'valid');
        
	    weight_x = w0 * max( abs(dx), thr_e).^( exp_a - 2);
	    weight_y = w0 * max( abs(dy), thr_e).^( exp_a - 2);
	    weight_xx = 0.25 * w0 *max( abs(dxx),thr_e).^(exp_a-2);
	    weight_yy = 0.25 * w0 *max( abs(dyy),thr_e).^(exp_a-2);
	    weight_xy = 0.25 * w0 *max( abs(dxy),thr_e).^(exp_a-2);
	end
	
	%deconvolve the image with the new weights
    if ~useValid && ( 0 == mod( t, 2 ) )
        continue;
    else
        [x] = deconvL2_wSegmentColor( L1_big, validL1_big, KL1, wk, hk, ...
    weight_x, weight_y, weight_xx, weight_yy, weight_xy, parameter);
    end
end
if doVisualize
    fprintf('\n' )
end
sharp = reshape( x, height, width, colors);
%remove the boundary that was added in the process
sharp = sharp( hk + (1:height0), wk + (1:width0), : );
