%%%%%%%%%%%%%%%%%%%
% Script to deblur an image given the Piecewise Rigid Scene Flow computed
% between stereo video frames
% as described in the paper: 
% Sellent, Rother, Roth: "Stereo Video Deblurring",
% In: ECCV 2016 (to appear)
%
% Implementation by Anita Sellent (anita.sellent@tu-dresden.de)
% 
% Computer Vision Lab Dresden
% Technische Universitaet Dresden, Germany
% 
% Last revision: July 2016
%
% This program can be used only for research purposes.
% This program is distributed WITHOUT ANY WARRANTY; 
% without even the implied warranty of MERCHANTABILITY 
% or FITNESS FOR A PARTICULAR PURPOSE.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

close all;
clear;

%% algorithm parameters
parameter.alpha = 0.001;%smoothness parameter of image
parameter.nIter = 25;%iterations for L2 deblurring subroutine
parameter.nIntegrationSteps = 70;
parameter.validWeight = 4000;%to update the validity masks
parameter.alterIters = 10 + 1;%iterate between smoothing weight and validity weigh update
%% switches
parameter.useValid = true;

parameter.use3DMotion = true;%either use 2d motion field ( optical flow ) or 3D proposals
parameter.useMatExp = true;%only relevant for 3D motion interpolation

parameter.doVisualize = true;%visualize intermediate results

iDeblur = 2;%for comparison with video deblurring, evaluate on second frame

%% test case
flowPath = fullfile( '..', 'results' );
imgPath = fullfile( '..', 'data' );
resultPath = 'results';

if ~exist( resultPath, 'dir' )
    mkdir( resultPath )
end

dataName = { ...
%     'apple_B'; ...
%     'bunny_B'; ...
   'chair_B'; ...
%     'squares_B'; ...
%     'triplane_B'...
   }

%camera settings
%We use camera settings to read out the actual ratio exposure time and 1/framerate
%for the rendered scenes this is 1
parameter.tau = 1;%maximal exposure time of 1 unit. In real cameras tau < 1. 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% load data
for iData = 1: length( dataName )
    display( dataName{iData} )
    %get image
    L1 = im2double( imread( fullfile( ...
        imgPath, [ dataName{iData} int2str( iDeblur ) '_cam0.png'] ) ) );
    
    [height, width, colors] = size( L1 );
    
    L10 = L1;
    
    %get flow and convert from the notation of 
    %to the notation used here
        sourceName = fullfile( flowPath, ...
            [dataName{iData} '_sceneFlowOut.mat'] );
        result = load( sourceName );

        [ fieldsL1, validL1, segs, Rt, normals, calib ] = ...
            convertDataSceneFlow( result );%fields contains disparity, X flow, Y flow stacked
    
    %deblur images
    [S_L1] = reconstructSharp3D( L10, segs, ...
	fieldsL1, ...
	Rt, normals, calib, ...
	validL1, parameter );
    
    %compare to ground truth sharp image
    refName = fullfile( ...
            imgPath, [ dataName{iData}(1:end-1) 'S' int2str( iDeblur ) '_cam0.png'] );
    if exist( refName, 'file' )
        J0 = im2double( imread( refName ) );

        % evaluation
        p = psnr( S_L1, J0 )
        %
        % http://www.ece.uwaterloo.ca/~z70wang/research/ssim/
        %s = ssim( mean( J0, 3)*255, mean( S_L1, 3)*255 )
    end
    
    %% save the final image
    saveName = fullfile( resultPath, [ dataName{iData} '_sharpL' int2str( iDeblur ) ] );
    imwrite( min( 1, max( 0, S_L1) ), [ saveName '.png' ] );
    
end

