//////////////////////////////
// Script to deblur an image given the Piecewise Rigid Scene Flow computed
// between stereo video frames
// as described in the paper: 
// Sellent, Rother, Roth: "Stereo Video Deblurring",
// In: ECCV 2016 (to appear)
//
// Implementation by Anita Sellent (anita.sellent@tu-dresden.de)
// 
// Computer Vision Lab Dresden
// Technische Universitaet Dresden, Germany
// 
// Last revision: July 2016
//
// This program can be used only for research purposes.
// This program is distributed WITHOUT ANY WARRANTY; 
// without even the implied warranty of MERCHANTABILITY 
// or FITNESS FOR A PARTICULAR PURPOSE.
//
////////////////////////////// 

#include "mex.h"
#include <stdio.h>
#include <math.h>
#include <opencv2/opencv.hpp>
#include <vector>


#define max(a,b) ((a)>(b)?(a):(b))
#define min(a,b) ((a)>(b)?(b):(a))

#define round(x) (x >= 0 ? int(x + 0.5) : int(x-0.5))

//[ row, col, val ] = mx_make_matrix_of_rigid_full( ...
//    Kl, Tl, frame, Rt, normals, seg, bbx0, bbx1, bby0, bby1, segId, step, ...
//    useSymmetric, useMatExp, useZeroBoundaries, useTransposedNotation );
void mexFunction(int nlhs,mxArray *out[],int nrhs,const mxArray *prhs[])
{
    cv::Mat Klt( 3, 3, CV_64F, mxGetPr(prhs[0]));//opencv: row major. Matlab: column major
    cv::Mat Tl( 3, 1, CV_64F, mxGetPr(prhs[1]));
    const int frame = mxGetScalar(prhs[2]);
    double* Rt = mxGetPr(prhs[3]);   
    double* normals = mxGetPr(prhs[4]);
    double* seg = mxGetPr(prhs[5]);
    const int bbx0 = mxGetScalar(prhs[6]);//bounding box of image pixels where matrix will be operating 
    const int bbx1 = mxGetScalar(prhs[7]);//in MATLAB indexing !
    const int bby0 = mxGetScalar(prhs[8]); 
    const int bby1 = mxGetScalar(prhs[9]); 
    double* segId = mxGetPr(prhs[10]);
    const int step = mxGetScalar(prhs[11]);
    const double time0 = mxGetScalar(prhs[12]);
    const double time1 = mxGetScalar(prhs[13]);
    bool useMatExp = mxGetScalar(prhs[14]);
    bool useZeroBoundaries = mxGetScalar(prhs[15]);
    bool useTransposedNotation = mxGetScalar(prhs[16]);
    
    //constants
    const int nSeg = mxGetM(prhs[10]);
    const int height0 = mxGetM(prhs[5]);
    const int width0 = mxGetN(prhs[5]);
    
    int height = bby1 - bby0 + 1;
    int width = bbx1 - bbx0 + 1;
    
    double normalization;
    normalization = 1.0 / ( 2.0 * step + 1.0 );
    
    double dt = ( time1 - time0 ) / ( 2.0 * step );
    
    int size[] = { height * width + 1, height * width + 1};//work in matlab notaion
    cv::SparseMat sparse_mat(2, size, CV_64F);
    
    //containers
    double tmpR[9]; 
    double tmpT[3];
    double tmpN[3];
    
    //iterate over segments
    for( int iSeg = 0; iSeg < nSeg; iSeg++ )
    {
      int currentSeg = segId[iSeg];
      //get the right formulation of the rigid motion
      for ( int i = 0; i < 3; i++ )
      {
	int a = 12 + i + currentSeg*16;
	tmpT[i] = Rt[a];
	for ( int j = 0; j < 3; j++ )
	{
	  int index = j + i*4 + currentSeg*16;
	  tmpR[ i + 3*j] = Rt[index];
	}
	int b = i + currentSeg * 3;
	tmpN[i] = normals[b];
      }
      cv::Mat R(3, 3, CV_64F, tmpR);
      cv::Mat T(3, 1, CV_64F, tmpT);
      cv::Mat n(3, 1, CV_64F, tmpN);

      double trace = R.at<double>(0,0) + R.at<double>(1,1) + R.at<double>(2,2);
      double theta = acos( 0.5*( trace - 1.0 ) );
      cv::Mat u;
      if ( fabs(theta) < 0.00001 )
      {
        useMatExp = false;//approximation is true for small rotations
        u = cv::Mat_<double>(3,1) << 0, 0, 1;
      }
      else
      {
        u = (cv::Mat_<double>(3,1) << 
	  R.at<double>(2,1) - R.at<double>(1,2), R.at<double>(0,2) - R.at<double>(2,0), R.at<double>(1,0) - R.at<double>(0,1) );
        u = u / ( 2.0*sin(theta) );
      }
      
      cv::Mat K;
      K = ( cv::Mat_<double>(3,3) << 
	  0, -u.at<double>(2) , u.at<double>(1),
	  u.at<double>(2), 0, -u.at<double>(0),
	  -u.at<double>(1), u.at<double>(0),0 );
      cv::Mat Ksquare = K * K;
      if ( 1 == frame )//%move normal to current frame
      {
        n = R * n;
        n = n / ( 1 - T.t()*n );
      }
      //projection into the image
      cv::Mat Pr = Klt.t() * ( cv::Mat::eye(3, 3, CV_64F) - Tl * n.t());

      //precalculations for matExp
      cv::Mat iAT;
      cv::Mat uut_theta;
      if ( true == useMatExp )
      {
	uut_theta = u * u.t() * theta;
	cv::Mat A = (1 - cos( theta ) ) * K - sin( theta ) * Ksquare + uut_theta;
        iAT = A.inv()  * T;
      }
      //iterate over time
      for( int it = 0; it <= 2*step; it++ )
      {
	double time = time0 + it * dt;
	
	cv::Mat Rc = cv::Mat::eye(3, 3, CV_64F) + 
		     sin( time * theta ) * K + 
		     ( 1 - cos( time * theta ) ) * Ksquare;
	cv::Mat tc;
        if ( true == useMatExp )
            tc = ( ( 1- cos( time * theta ) ) * K - sin( time * theta ) * Ksquare + uut_theta * time ) * iAT;
        else
            tc = time * T;
 
	cv::Mat n0c = Rc*n;
        n0c = n0c / ( 1 - tc.t() * n );
	//backproject for current time
        cv::Mat mat_l0 = ( Klt.t() * ( cv::Mat::eye(3, 3, CV_64F) - Tl * n0c.t() ) ).inv();
        //move to position at t = 0
        cv::Mat tmp1 = ( Rc - tc * n.t() ).inv();
        mat_l0 = tmp1 * mat_l0;
        //project back into the image
        mat_l0 = Pr * mat_l0;

	//iterate over image coordinates (MATLAB indexing)
	for( int x = bbx0; x <= bbx1; x++ )
	{
	  int xx = min( max( x, 1), width0 );
	  for ( int y = bby0; y <= bby1; y++ )
	  {
	    int yy = min( max( y, 1), height0 );
	    int indx = ( (xx-1)*height0 + yy ) - 1;//transform matlab coordinates into c index
	    if ( seg[ indx] == currentSeg )
	    {
	      //start point
	      int tmp_row;
	      if ( false == useTransposedNotation )
		tmp_row = (y - bby0 + 1) + height * ( x - bbx0 );//transform matlab coordinates into matlab index
	      else
		tmp_row = ( y - bby0 ) * width + ( x - bbx0 + 1);//transform matlab coordinates into matlab index

	      //image coordinates
	      cv::Mat Pl0 = mat_l0 * ( cv::Mat_<double>(3,1) << x, y, 1 );
	      double Xnew = Pl0.at<double>(0) / Pl0.at<double>(2);
	      double Ynew = Pl0.at<double>(1) / Pl0.at<double>(2);
	      //transform into new image - from here only use this image coordinates!
	      Xnew = Xnew - bbx0 + 1;
	      Ynew = Ynew - bby0 + 1;
	      //boundaries
	      if ( ( true == useZeroBoundaries ) && 
		( ( 1 > Xnew ) || ( width < Xnew ) || ( 1 > Ynew ) || ( height < Ynew ) ) )
	      {
		continue;//zero values do not need to be specified in a sparse matrix
	      }
	      Xnew = max( 1, min( width, Xnew ) );
	      Ynew = max( 1, min( height, Ynew ) );
        
	      //bilinear interpolation
	      int X0 = floor( Xnew );
	      int Y0 = floor( Ynew );
	      int X1 = min( X0 + 1, width );
	      int Y1 = min( Y0 + 1, height );
	      double xi = ( Xnew - X0 );
	      double yi = ( Ynew - Y0 );
	      
	      double val;
	      int row;
	      int col;
	      
	      //the upper left corner
	      val = ( ( 1 - xi ) * ( 1 - yi ) * normalization );
	      row = ( tmp_row );
	      if ( useTransposedNotation )
		  col = ( ( Y0 - 1) * width + X0 );
	      else
		  col = ( ( X0 - 1) * height + Y0 );
	      sparse_mat.ref<double>( row, col ) += val;
 
	      //the lower left corner
	      val = ( ( 1 - xi ) * yi * normalization);
	      row = ( tmp_row );
	      if ( useTransposedNotation )
		  col = ( ( Y1 - 1) * width + X0 );
	      else
		  col = ( ( X0 - 1) * height + Y1 );
	      sparse_mat.ref<double>( row, col ) += val;

	      //the upper right corner
	      val = ( xi * ( 1 - yi) * normalization );
	      row = ( tmp_row );
	      if ( useTransposedNotation )
		  col = ( ( Y0 - 1) * width + X1 );
	      else
		  col = ( ( X1 - 1) * height + Y0 );
	      sparse_mat.ref<double>( row, col ) += val;
	      
	      //the lower right corner
	      val = ( xi * yi * normalization );
	      row = ( tmp_row );
	      if ( useTransposedNotation )
		  col = ( ( Y1 - 1) * width + X1 );
	      else
		  col = ( ( X1 - 1) * height + Y1 );
	      sparse_mat.ref<double>( row, col ) += val;
	    }// if seg
	  }//y
	}//x
      }//time
    }//iSeg
    
    //transform to vectors
    std::vector<int> row;
    std::vector<int> col;
    std::vector<double> val;
    
    cv::SparseMatConstIterator_<double> it = sparse_mat.begin<double>(), it_end = sparse_mat.end<double>();

    for(; it != it_end; ++it)
    {
      const cv::SparseMat::Node* n = it.node();
      row.push_back( n->idx[0] );
      col.push_back( n->idx[1] );
      val.push_back( it.value<double>() );
    }

    //output
    int nElems = val.size();
    out[0] = mxCreateDoubleMatrix(nElems, 1, mxREAL);
    out[1] = mxCreateDoubleMatrix(nElems, 1, mxREAL); 
    out[2] = mxCreateDoubleMatrix(nElems, 1, mxREAL);
    
    double* prow = mxGetPr(out[0]);
    double* pcol = mxGetPr(out[1]);
    double* pval = mxGetPr(out[2]);
    
    for ( int i = 0; i < nElems; i++ )
    {
      prow[i] = row.at( i );
      pcol[i] = col.at( i );
      pval[i] = val.at( i );
    }

}